//Demo code
//Doug Densmore
//Will Plishker

extern "C" {
#include <conio.h>
#include <unistd.h>
#include <dsound.h>
#include <dmotor.h>
#include <time.h>
#include <tm.h>
#include <dsensor.h>
}
#include <c++/sensor.h>

#define TURN_NUM 270
#define TURN_DIS_RIGHT 143
#define TURN_DIS_LEFT 132
#define DEFAULT_SPEED 60
#define TURN_SPEED 38
#define FWD_DIS 175
#define MAX_Y 4
#define MAX_X 5
#define TURN180 982
#define BLOCK_FOUND 58
#define TIMER 10

typedef enum {
    call,
    schedule,
    future,
}Opcode;

typedef enum {
    forward,
    right,
    left,
    backward,
    stopped,
}Action_State;

typedef enum {
    start,
    turning,
    moving,
}Turn_State;

typedef enum {
    left_turn,
    right_turn,
    no_turn,
}Turn_Dir;

typedef struct {
    void(*(fp))();
    int priority;
}Task;

typedef struct {
    Opcode opcode;
    int index;
    void(*driver)();
    Task task;
}Ecode;

typedef struct {
    int done_moving;
    int block_found;
}NavigateInput_t;

typedef struct {
    Action_State action_state;
    int block_found;
    int block_up;
}MoveInput_t;

void lcd_driver(void);
void beep_driver(void);
void beep_task(void);
void lcd_task(void);
void rotation_driver(void);
void motor_driver(void);
void navigation_task(void);
void move_task(void);
void intertask_driver(void);
void light_driver(void);

volatile int block_found = 0;
extern volatile int rtos_timer;
Ecode ecode[100];
int lcd_in = 0;
note_t beep_in[2];
Task scheduled_tasks[100];
int head_scheduled_tasks=0, rot_sensor_axel=0, rot_sensor_dir=0, moveOutput;
int axel_motor_speed=0, turn_motor_speed=0, arm_motor_speed = 0;
Turn_Dir turn_motor_dir=no_turn;
MotorDirection axel_motor_dir=off, arm_motor_dir=off;
int reset_rot_sensor_axel = 0;
int x=0, y=0;
NavigateInput_t navigateInput;
MoveInput_t moveInput;
Action_State navigateOutput;
LightSensor *light;

void execute();

void initialize(){
    rtos_timer=0;
    ds_init();
    ds_active(&SENSOR_1);
    ds_active(&SENSOR_2);
    light = new LightSensor(&SENSOR_3, 1);
    ds_rotation_on(&SENSOR_1);
    ds_rotation_on(&SENSOR_2);

    motor_a_dir(off);
    motor_a_speed(DEFAULT_SPEED);

    motor_b_dir(off);
    motor_b_speed(DEFAULT_SPEED);

    cputw(light->value());
    //Raise arm
    motor_c_dir(fwd);
    motor_c_speed(MAX_SPEED);
    while(light->value()<200)
        cputw(light->value());
    motor_c_dir(off);
    sleep(1);

    motor_b_dir(fwd);
    motor_b_speed(MAX_SPEED);
    while(light->value()<200)
        cputw(light->value());

    motor_b_dir(off);
    motor_b_speed(0);
    ds_rotation_set(&SENSOR_1, 0);
    ds_rotation_set(&SENSOR_2, 0);

    motor_b_dir(fwd);
    motor_b_speed(DEFAULT_SPEED);
    while(ROTATION_2 < TURN180)
        cputw(ROTATION_2);
    motor_b_dir(off);
    
    ds_rotation_set(&SENSOR_2, 0);

    beep_in[0].pitch = PITCH_E3; beep_in[0].length = EIGHTH;
    beep_in[1].pitch = PITCH_END;beep_in[1].length = 0;

    ecode[0].opcode = call;     ecode[0].driver  = beep_driver;
    ecode[1].opcode = call;     ecode[1].driver  = lcd_driver;
    ecode[2].opcode = call;     ecode[2].driver  = rotation_driver;
    ecode[3].opcode = call;     ecode[3].driver  = motor_driver;
    ecode[4].opcode = call;     ecode[4].driver  = intertask_driver;
    ecode[5].opcode = call;     ecode[5].driver  = light_driver;
    ecode[6].opcode = schedule; ecode[6].task.fp = beep_task;       ecode[6].task.priority=2;
    ecode[7].opcode = schedule; ecode[7].task.fp = lcd_task;        ecode[7].task.priority=1;
    ecode[8].opcode = schedule; ecode[8].task.fp = navigation_task; ecode[8].task.priority=1;
    ecode[9].opcode = schedule; ecode[9].task.fp = move_task;       ecode[9].task.priority=1;
    ecode[10].opcode = future;  ecode[10].index  = 20;

    ecode[20].opcode = call;     ecode[20].driver  = rotation_driver;
    ecode[21].opcode = call;     ecode[21].driver  = lcd_driver;
    ecode[22].opcode = call;     ecode[22].driver  = motor_driver;
    ecode[23].opcode = call;     ecode[23].driver  = intertask_driver;
    ecode[24].opcode = call;     ecode[24].driver  = light_driver;
    ecode[25].opcode = schedule; ecode[25].task.fp = lcd_task;        ecode[25].task.priority=1;
    ecode[26].opcode = schedule; ecode[26].task.fp = move_task;       ecode[26].task.priority=1;
    ecode[27].opcode = future;   ecode[27].index   = 0;

    navigateOutput = stopped;
    moveOutput     = 1; //done_moving
    moveInput.block_up = 0;
}//end initialize

void intertask_driver(){
    moveInput.action_state = navigateOutput;
    navigateInput.done_moving = moveOutput;
}

void beep_driver(){
    dsound_play(beep_in);
    wait_event(dsound_finished,0);
 
}//end beep

void lcd_driver(){
    cputw(lcd_in);
}

void light_driver(){
    if((light->value()<BLOCK_FOUND))
        navigateInput.block_found = 1;
    else if(!navigateInput.block_found)
        navigateInput.block_found = 0;

    moveInput.block_found = navigateInput.block_found;

    if((moveInput.block_found)&&(light->value()>150))
        moveInput.block_up = 1;
}

void rotation_driver(){
    rot_sensor_axel = ROTATION_1;
    rot_sensor_dir  = ROTATION_2;

    if(reset_rot_sensor_axel)
        ds_rotation_set(&SENSOR_1, 0);

    reset_rot_sensor_axel = 0;
}

void motor_driver(){    //move forward
    motor_a_dir(axel_motor_dir);
    motor_a_speed(axel_motor_speed);

    //turn motor
    switch(turn_motor_dir){
    case left_turn:  
        motor_b_dir(fwd);
        motor_b_speed(turn_motor_speed);
        break;

    case right_turn: 
        motor_b_dir(rev);
        motor_b_speed(turn_motor_speed);
        break;

    case no_turn:
        motor_b_dir(off);
        motor_b_speed(turn_motor_speed);
        break;
    }

    motor_c_dir(arm_motor_dir);
    motor_c_speed(arm_motor_speed);
}


void lcd_task(){
    lcd_in = (x<<8)|(y);
    //    lcd_in = rot_sensor_axel;
    //lcd_in = light->value();
}

void beep_task(){
    static note_t ourmusic[11];
    static int i=0;

    ourmusic[0].pitch = PITCH_E3;
    ourmusic[0].length = EIGHTH;
    ourmusic[1].pitch = PITCH_D3;
    ourmusic[1].length = EIGHTH;
    ourmusic[2].pitch = PITCH_C3;
    ourmusic[2].length = EIGHTH;
    ourmusic[3].pitch = PITCH_D3;
    ourmusic[3].length = EIGHTH;
    ourmusic[4].pitch = PITCH_E3;
    ourmusic[4].length = EIGHTH;
    ourmusic[5].pitch = PITCH_E3;
    ourmusic[5].length = EIGHTH;
    ourmusic[6].pitch = PITCH_E3;
    ourmusic[6].length = QUARTER;
    ourmusic[7].pitch = PITCH_D3;
    ourmusic[7].length = EIGHTH;
    ourmusic[8].pitch = PITCH_D3;
    ourmusic[8].length = EIGHTH;
    ourmusic[9].pitch = PITCH_D3;
    ourmusic[9].length = QUARTER;
    ourmusic[10].pitch = PITCH_END;
    
    beep_in[0].pitch=ourmusic[i].pitch;
    i++;
    if(i>10) i=0;


}

int main(int argc, char **argv) {
    int i=0;
    cputs("Hello");
    sleep(1);
    cputs("EE290O");
    sleep(1);
    cputs("Demo");
    sleep(1);
    cls("     ");
    initialize();

    while(1) {
        if(rtos_timer>TIMER) {
            rtos_timer= 0;
            head_scheduled_tasks = 0;
            while(ecode[i].opcode!=future) {
                if(ecode[i].opcode==call)
                    (*(ecode[i].driver))();

                if(ecode[i].opcode==schedule){
                    scheduled_tasks[head_scheduled_tasks] = ecode[i].task;
                    head_scheduled_tasks++;
                }
                i++;
            }
            i = ecode[i].index;
            execute();
        }
    }
    return 0;
}

void move_task() {
    static Turn_State turn_state = start;
    static int done_acked = 1;
    static int do_once=0;

    if((moveInput.block_found)&&(!moveInput.block_up)) {
        arm_motor_dir = rev;
        arm_motor_speed = MAX_SPEED;
        axel_motor_dir = off;
        turn_motor_dir = no_turn;
        do_once = 0;
        return;
    } else if((moveInput.block_found) && (!do_once)) {
        turn_state = start;
        arm_motor_dir = brake;
        do_once = 1;
    }

    if(done_acked) { //if the done has been acknowledged
        //cputw(0xAAAA);
        moveOutput = 0; //accept next command by saying not done
        switch(moveInput.action_state) {
        case right:
            switch(turn_state) {
            case start: 
                if(turn_motor_dir > (-TURN_NUM+30)){
                    turn_motor_dir = right_turn;
                    turn_motor_speed = DEFAULT_SPEED;
                }
                turn_state = turning;
                break;
            case turning:
                if(rot_sensor_dir < (-TURN_NUM)){
                    turn_motor_dir = no_turn;
                    turn_motor_speed = MIN_SPEED;
                    turn_state = moving;
                    axel_motor_dir = rev;
                    axel_motor_speed = DEFAULT_SPEED;
                }
                break;
            case moving:
                if(rot_sensor_axel < (-TURN_DIS_RIGHT)) {
                    axel_motor_dir = off;
                    axel_motor_speed = TURN_SPEED;
                    reset_rot_sensor_axel = 1;
                    moveOutput = 1;
                    turn_state = start;
                    done_acked = 0;
                }
            }
            return;
        case left:
            switch(turn_state) {
            case start: 
                if(rot_sensor_dir < (TURN_NUM - 30)) {
                    turn_motor_dir = left_turn;
                    turn_motor_speed = DEFAULT_SPEED;
                    moveOutput = 0;
                }
                turn_state = turning;
                break;
            case turning:
                if(rot_sensor_dir > (TURN_NUM)) {
                    turn_motor_dir = no_turn;
                    turn_motor_speed = MIN_SPEED;
                    turn_state = moving;
                    axel_motor_dir = rev;
                    axel_motor_speed = TURN_SPEED;
                }
                break;
            case moving:
                if(rot_sensor_axel < (-TURN_DIS_LEFT)) {
                    axel_motor_dir = off;
                    axel_motor_speed = MIN_SPEED;
                    reset_rot_sensor_axel = 1;
                    turn_state = start;
                    moveOutput = 1;
                    done_acked = 0;
                }
                break;
            }
            return;
        case forward:
            //cputw(0xBBBB);
            switch(turn_state) {
            case start: 
                //cputw(0xCCCC);
                if(rot_sensor_dir < -30)
                    turn_motor_dir = left_turn;
                if(rot_sensor_dir > 30)
                    turn_motor_dir = right_turn;
                turn_motor_speed = DEFAULT_SPEED;
                turn_state = turning;
                break;
            case turning:
                //cputw(0xDDDD);
                if((turn_motor_dir == left_turn)&&(rot_sensor_dir < -2))
                    {} 
                else {
                    if((turn_motor_dir == right_turn)&&(rot_sensor_dir > 2))
                        {} 
                    else {
                        turn_motor_dir = no_turn;
                        turn_motor_speed = MIN_SPEED;
                        turn_state = moving;
                        axel_motor_dir = rev;
                        axel_motor_speed = DEFAULT_SPEED;
                    }
                }
                break;
            case moving:
                //                cputw(0xEEEE);
                if(rot_sensor_axel < (-FWD_DIS)) {
                    axel_motor_dir = off;
                    axel_motor_speed = MIN_SPEED;
                    reset_rot_sensor_axel = 1;
                    moveOutput = 1;
                    turn_state = start;
                    done_acked = 0;
                }
            }
            return;
        case stopped:
            turn_state = start;
            return;
        default:
            turn_state = start;
            return;
        }
    } else { //else navigate hasn't realized we're done
        if(moveInput.action_state == stopped)   //if navigate realized we're done
            done_acked = 1;        // we are ready for the next command
    }
}

void navigation_task () {
    static Action_State action_state = stopped;

    if((action_state==stopped)) {
        static Action_State dir = forward;
        if(!navigateInput.block_found) {
            if((MAX_X+1)==x) {
                action_state = stopped;
            } else if(((y==MAX_Y)&&(dir==forward))){
                action_state = right;
                dir = right;
                x++;
            } else if(((y==MAX_Y)&&(dir==right))) {
                action_state = right;
                dir = backward;
                y--;
            } else if(((y==0)&&(dir==backward))) {
                action_state = left;
                dir = right;
                x++;
            } else if(((y==0)&&(dir==right))) {
                action_state = left;
                dir = forward;
                y++;
            } else {
                action_state = forward;
                if(dir==forward)
                    y++;
                else
                    y--;
            }
        } else {
            switch(dir) {
            case forward: 
                action_state = left;
                dir = left;
                x--;
                break;
            case right:
                action_state = right;
                dir = backward;
                y--;
                break;
            case left:
                if(x>0){
                    action_state = forward;
                    dir = left;
                    x--;
                } else if(y>0) {
                    action_state = left;
                    dir = backward;
                    y--;
                } else {
                    action_state = stopped;
                }
                break;
            case backward:
                if(y>0){
                    action_state = forward;
                    dir = backward;
                    y--;
                } else if(x>0) {
                    action_state = right;
                    dir = left;
                    x--;
                } else {
                    action_state = stopped;
                }
                break;
            }
        }
    } else if(!navigateInput.done_moving) {
        navigateOutput = action_state;
    } else { //else ack the end of move
        action_state = stopped;
        navigateOutput = action_state;
    }

    return;
}



void execute() {
    int i, current;

    for(i=0; i<10; i++){
        current = head_scheduled_tasks;
        while(current>0) {
            current--;
            if(scheduled_tasks[current].priority==i) 
                scheduled_tasks[current].fp();
        }
    }
}


    /* ourmusic2[0].pitch = PITCH_E3;
       ourmusic2[0].length = EIGHTH;
       ourmusic2[1].pitch = PITCH_G3;
       ourmusic2[1].length = EIGHTH;
       ourmusic2[2].pitch = PITCH_G3;
       ourmusic2[2].length = QUARTER;

       ourmusic2[3].pitch = PITCH_E3;
       ourmusic2[3].length = EIGHTH;
       ourmusic2[4].pitch = PITCH_D3;
       ourmusic2[4].length = EIGHTH;
       ourmusic2[5].pitch = PITCH_C3;
       ourmusic2[5].length = EIGHTH;
       ourmusic2[6].pitch = PITCH_D3;
       ourmusic2[6].length = EIGHTH;
       ourmusic2[7].pitch = PITCH_E3;
       ourmusic2[7].length = EIGHTH;
       ourmusic2[8].pitch = PITCH_E3;
       ourmusic2[8].length = EIGHTH;
       ourmusic2[9].pitch = PITCH_E3;
       ourmusic2[9].length = QUARTER;
       ourmusic2[10].pitch = PITCH_END;

       ourmusic3[0].pitch = PITCH_E3;
       ourmusic3[0].length = EIGHTH;
       ourmusic3[1].pitch = PITCH_D3;
       ourmusic3[1].length = EIGHTH;
       ourmusic3[2].pitch = PITCH_D3;
       ourmusic3[2].length = EIGHTH;
       ourmusic3[3].pitch = PITCH_E3;
       ourmusic3[3].length = EIGHTH;
       ourmusic3[4].pitch = PITCH_D3;
       ourmusic3[4].length = EIGHTH;
       ourmusic3[5].pitch = PITCH_C3;
       ourmusic3[5].length = QUARTER;
    */

